#!/usr/bin/perl -w

# $Id: webwatch_2.pl,v 1.6 2020/04/10 22:52:05 pmdg Exp $

#########################################################################
# webwatch_2.pl
#
# basic web version of 'nsrwatch'. Second version eschews passing info
# via datafiles and drops the volumes query. We query within the CGI
# itself.
#
# Key REST API elements we need to monitor:
#
# === ALERTS ===
# /nwrestapi/v3/global/alerts
#		count: X
#		alerts: list of hashes
#			category:	string
#			message:	string
#			priority:	string
#			timestamp: 	string
#
#
# === SERVER MESSAGES ===
# /nwrestapi/v1/global/servermessages
#		count: X
#		serverMessages: list of hashes
#			category: text -> ucfirst per word
#			id: zero padded message ID, sort by this desc.
#			message: text
#			priority: text
#			source: text
#			timestamp: NW timestamp format.
# 
#
# === DEVICES ===
# /nwrestapi/v3/global/devices
#		count: X
#		devices: list of hashes
#			name
#			mediaFamily
#			mediaType
#			readOnly ("true" or "false")
#			status ("Enabled" or other)
#			volumeId
#			volumePool
#
#
# === SESSIONS ===
# NB: Not just active sessions.
# /nwrestapi/v3/global/sessions
#		count: X
#		sessions: list of hashes
#			completed: true or false
#			clientHostname
#			protectionGroup
#			device
#			mode: Saving, or others
#			pool
#			saveSet
#			size -> hash -> { unit, value }
#			startTime
#			stopped: true or false
#			transferRate -> hash -> { unit, value }
#			volume
#
#########################################################################

#########################################################################
# Modules
#########################################################################
use strict;
use JSON;
use Data::Dumper;
use REST::Client;
use MIME::Base64;
use CGI;
use File::Basename;


#########################################################################
# Global Setup
#########################################################################

# Turn debug on for logs written to the web server logs.
my $debug = 1;

# Create a "monitor" user in NetWorker and setup the credentials in the
# username and password below.
my $user = "watcher";
my $pass = "Backup321_";

# Change the server to reference your NetWorker Server.
my $server = "orilla.turbamentis.int";
my $auth = encode_base64($user . ":" . $pass);

# If you are using self-signed ceriticates, you'll need to keep
# the following line enabled.
$ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = 0;


#########################################################################
# Global Variables - Should be no need to change these.
#########################################################################

my %data = ();
my $headers = { Accept => "application/json", 
		 	    Authorization => "Basic $auth"
		 	  };
my $self = basename($0);


#########################################################################
# Subroutines
#########################################################################

# in_list($expr,@list) returns true iff $expr is in list
sub in_list {
	return 0 if (@_+0 < 2);
	
	my $element = $_[0];
	return 0 if (!defined($element));
	shift @_;
	
	my @list = @_;
	return 0 if (!@_ || @_+0 == 0);
	
	$element = quotemeta($element);
	
	my @results = grep(/^$element$/,@list);
	return @results+0;
}

# in_gb($value,$unit) returns the given NetWorker unit combination in GB.
sub in_gb {
	return 0 if (@_+0 != 2);
	
	my $value = $_[0];
	my $unit = lc($_[1]);
	
	return $value if ($unit eq "gb");
	
	# else...
	if ($unit eq "b") {
		return $value / 1000 / 1000 / 1000;
	} elsif ($unit eq "kb") {
		return $value / 1000 / 1000;
	} elsif ($unit eq "mb") {
		return $value / 1000;
	} elsif ($unit eq "tb") {
		return $value * 1000;
	} elsif ($unit eq "pb") {
		return $value * 1000 * 1000;
	}
}

# in_mb($value,$unit) returns the given NetWorker transfer unit combination in MB.
sub in_mb {
	return 0 if (@_+0 != 2);
	
	my $value = $_[0];
	my $unit = lc($_[1]);
	
	return $value if ($unit eq "gb");
	
	# else...
	if ($unit eq "b/s") {
		return $value / 1000 / 1000;
	} elsif ($unit eq "kb/s") {
		return $value / 1000;
	} elsif ($unit eq "mb/s") {
		return $value;
	} elsif ($unit eq "gb/s") {
		return $value * 1000;
	} elsif ($unit eq "tb/s") {
		return $value * 1000 * 1000;
	}
}

# clean_timestamp($expr) tries to clean up a timestamp for us.
sub clean_timestamp {
	my $timestamp = $_[0];
	
	my $date = "";
	my $time = "";
	if ($timestamp =~ /^(.*)T(.*)$/) {
		$date = $1;
		$time = $2;
		if ($time =~ /^(.*)[+-].*$/) {
			$time = $1;
		}
		return "$date $time";
	} else {
		# If we can't find a timestamp, return what we were passed.
		return $timestamp;
	}
}

# get_message_data() retrieves server messages and populates the hash.
sub get_message_data {
	my $client = REST::Client->new();
	$client->setHost("https://$server:9090");
	$client->GET('/nwrestapi/v1/global/servermessages', $headers);
	
	$debug && print STDERR "$self Query $server for messages.\n";
	my $response = from_json($client->responseContent());
	my %messages = %{$response};
	
	if (defined($messages{"count"})) {
		$debug && print STDERR "$self Processing " . $messages{"count"} . " messages.\n";
	}
	
	$data{messages}{count} = $messages{"count"};
	if (defined($messages{"serverMessages"})) {
		foreach my $mess (@{$messages{"serverMessages"}}) {
			my %message = %$mess;
			
			# This is a simple copy & paste into our data structure.
			my $messageID = $message{"id"};
			$data{messages}{message}{$messageID}{category} = $message{"category"};
			$data{messages}{message}{$messageID}{priority} = $message{"priority"};
			$data{messages}{message}{$messageID}{source} = $message{"source"};
			$data{messages}{message}{$messageID}{timestamp} = clean_timestamp($message{"timestamp"});
			$data{messages}{message}{$messageID}{message} = $message{"message"};
		}
	}
}

# get_session_data() retrieves active session data and populates the hash.
sub get_session_data {
	my $client = REST::Client->new();
	$client->setHost("https://$server:9090");
	$client->GET('/nwrestapi/v3/global/sessions',$headers);
	
	$debug && print STDERR "$self Query $server for session details.\n";
	my $response = from_json($client->responseContent());

	my %sessions = %{$response};
	if (defined($sessions{"count"})) {
		$debug && print STDERR "$self Processing " . $sessions{"count"} . " recorded sessions.\n";
	}

	my @sessionList = ();
	my $activeSessionCount = 0;
	if (defined($sessions{"sessions"})) {
		foreach my $sess (@{$sessions{"sessions"}}) {
			my %session = %$sess;
		
			my $completed = (defined($session{"completed"})) ? $session{"completed"} : -1;
			my $device = (defined($session{"device"})) ? $session{"device"} : "no device";
			my $mode = (defined($session{"mode"})) ? $session{"mode"} : "null";
			my $pool = (defined($session{"pool"})) ? $session{"pool"} : "null";
			my $saveset = (defined($session{"saveSet"})) ? $session{"saveSet"} : "no saveset";
			my $volume = (defined($session{"volume"})) ? $session{"volume"} : "null";
			my $startTime = (defined($session{"startTime"})) ? $session{"startTime"} : "null";
			my $stopped = (defined($session{"stopped"})) ? $session{"stopped"} : 0;
			my $client = (defined($session{"clientHostname"})) ? $session{"clientHostname"} : "";
			my $group = (defined($session{"protectionGroup"})) ? $session{"protectionGroup"} : "";
			my $sizeGB = 0;
			my $xferMB = 0;
		
			next if ($completed);	 # For the time being we ignore sessions that are complete.
		
			# else...
			$activeSessionCount++;
			$completed = "Not Completed";
		
			if (defined($session{"size"})) {
				my %sz = %{$session{"size"}};
				my $sV = $sz{"value"};
				my $sU = $sz{"unit"};
				$sizeGB = in_gb($sV,$sU);
			}
			if (defined($session{"transferRate"})) {
				my %xfer = %{$session{"transferRate"}};
				my $xV = $xfer{"value"};
				my $xU = $xfer{"unit"};
				$xferMB = in_mb($xV,$xU);
			}
            $data{sessions}{session}{$activeSessionCount}{completed} = $completed;
            $data{sessions}{session}{$activeSessionCount}{mode} = $mode;
            $data{sessions}{session}{$activeSessionCount}{client} = $client;
            $data{sessions}{session}{$activeSessionCount}{group} = $group;
            $data{sessions}{session}{$activeSessionCount}{saveset} = $saveset;
            $data{sessions}{session}{$activeSessionCount}{device} = $device;
            $data{sessions}{session}{$activeSessionCount}{volume} = $volume;
            $data{sessions}{session}{$activeSessionCount}{start} = clean_timestamp($startTime);
            $data{sessions}{session}{$activeSessionCount}{stopped} = $stopped;
            $data{sessions}{session}{$activeSessionCount}{xferMB} = $xferMB;
            $data{sessions}{session}{$activeSessionCount}{sizeGB} = $sizeGB;
		}
	}
	$data{sessions}{count} = $activeSessionCount;
}

# get_alert_data() retrieves details of any active alerts and builds out the hash structure.
sub get_alert_data {
	$debug && print STDERR "$self Query $server for alert details.\n";
	my $client = REST::Client->new();
	$client->setHost("https://$server:9090");
	$client->GET(
		'/nwrestapi/v3/global/alerts',
		$headers);
		
	my $response = from_json($client->responseContent());
	my %alerts = %{$response};
	if (defined($alerts{"count"})) {
		$debug && print STDERR "$self Processing " . $alerts{"count"} . " alerts\n";
	}
	
	my $alertCount = 0;
	$data{alerts}{count} = $alerts{"count"};
	if (defined($alerts{"alerts"})) {
		foreach my $alert (@{$alerts{"alerts"}}) {
			my %al = %$alert;
			
			# Alert data can be simply copied.
			$data{alerts}{alert}{$alertCount}{category} = $al{"category"};
			$data{alerts}{alert}{$alertCount}{priority} = $al{"priority"};
			$data{alerts}{alert}{$alertCount}{timestamp} = clean_timestamp($al{"timestamp"});
			$data{alerts}{alert}{$alertCount}{message} = $al{"message"};
			
			$alertCount++;
		}
	}
}

# get_device_data() retrieves device data and builds out the hash structure for devices.
sub get_device_data {
	$debug && print STDERR "$self Query $server for device details.\n";
	my $client = REST::Client->new();
	$client->setHost("https://$server:9090");
	$client->GET(
		'/nwrestapi/v3/global/devices',
		$headers);
	
	my $response = from_json($client->responseContent());

	my %devices = %{$response};
	if (defined($devices{"count"})) {
		$debug && print STDERR "$self Processing " . $devices{"count"} . " devices\n";
	}
	
	my $deviceCount=0;
	$data{devices}{count} = $devices{"count"};
	if (defined($devices{"devices"})) {
		foreach my $device (@{$devices{"devices"}}) {
			my %dev = %$device;
		
			my $devName = (defined($dev{"name"})) ? $dev{"name"} : "null";
			my $mediaFamily = (defined($dev{"mediaFamily"})) ? $dev{"mediaFamily"} : "null";
			my $mediaType = (defined($dev{"mediaType"})) ? $dev{"mediaType"} : "null";
			my $readOnly = (defined($dev{"readOnly"})) ? $dev{"readOnly"} : "null";
			my $mountedVolume = (defined($dev{"mountedVolume"})) ? $dev{"mountedVolume"} : "None";
			my $lastMessage = (defined($dev{"message"})) ? $dev{"message"} : "<I>No activity</I>";
			my $status = (defined($dev{"status"})) ? $dev{"status"} : "null";
			my $volID = (defined($dev{"volumeId"})) ? $dev{"volumeId"} : "null";
			my $volPool = (defined($dev{"volumePool"})) ? $dev{"volumePool"} : "null";
			
			$data{devices}{device}{$devName}{family} = $mediaFamily;
            $data{devices}{device}{$devName}{type} = $mediaType;
            $data{devices}{device}{$devName}{readonly} = $readOnly;
            $data{devices}{device}{$devName}{volid} = $volID;
            $data{devices}{device}{$devName}{pool} = $volPool;
            $data{devices}{device}{$devName}{volume} = $mountedVolume;
            $data{devices}{device}{$devName}{lastmessage} = $lastMessage;
		}
	}
}

# get_data() calls the various data gathering functions.
sub get_data {
	# Just undef this in case at a later point we refresh ourselves dynamically without
	# a base program re-load.
	undef %data;
	
	get_device_data();
	get_session_data();
	get_alert_data();
	get_message_data();
}

# display_data() shows the data
sub display_data {
	print "<DIV class=all>\n";
	my $indent = "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;";
	
	print "<DIV class=sessions>\n";
	print "<H2> Sessions</H2>\n";
	if (%data && defined($data{sessions}) && defined($data{sessions}{count}) &&
		$data{sessions}{count} > 0) {
		print "<TABLE WIDTH=100% BORDER=0>\n";
		foreach my $sessionCount (keys $data{sessions}{session}) {
			my $client = $data{sessions}{session}{$sessionCount}{client};
			my $mode = $data{sessions}{session}{$sessionCount}{mode};
			my $saveset = $data{sessions}{session}{$sessionCount}{saveset};
			my $group = $data{sessions}{session}{$sessionCount}{group};
			my $device = $data{sessions}{session}{$sessionCount}{device};
			my $volume = $data{sessions}{session}{$sessionCount}{volume};
			my $startTime = $data{sessions}{session}{$sessionCount}{start};
			my $xferMB = $data{sessions}{session}{$sessionCount}{xferMB};
			my $sizeGB = $data{sessions}{session}{$sessionCount}{sizeGB};
			print "<TR>\n";
			print "<TD BGCOLOR=#C9C9C9><I>$mode</I></tTD>\n";
			print "<TD BGCOLOR=#C9C9C9><b>$client</b>";
			if ($saveset ne "" && $saveset ne "null") {
				print ":$saveset</TD>\n";
			} else {
				print "</TD>\n";
			}
			$xferMB = sprintf("%.3f",$xferMB);
			$sizeGB = sprintf("%.3f",$sizeGB);
			print "<TD BGCOLOR=#C9C9C9>$device:$volume</TD>\n";
			print "<TD BGCOLOR=#C9C9C9>$xferMB MB/s</TD>\n";
			print "<TD BGCOLOR=#C9C9C9>$sizeGB GB</TD>\n";
			print "</TR>\n";
			print "<TR>\n";
			print "<TD COLSPAN=3 BGCOLOR=#FAFAFA>$group</TD>\n";
			print "<TD COLSPAN=2 BGCOLOR=#FAFAFA>$startTime</TD>\n";
			print "</TR>\n";
		}
		
		print "</TABLE>\n";
	} else {
		print "<p><B><I>No active sessions</I></B>\n";
	}
	
	print "</DIV>\n"; 	# Sessions
	
	print "<DIV class=devices>\n";
	print "<H2> Devices</H2>\n";
	if (%data && defined($data{devices}) && defined($data{devices}{count}) &&
		$data{devices}{count} > 0) {
		print "<TABLE WIDTH=100% BORDER=0>\n";
		foreach my $deviceName (sort {lc $a cmp lc $b} keys $data{devices}{device}) {
			print "<TR>\n";
			my $family = $data{devices}{device}{$deviceName}{family};
			my $type = $data{devices}{device}{$deviceName}{type};
			my $readonly = $data{devices}{device}{$deviceName}{readonly};
			my $pool = $data{devices}{device}{$deviceName}{pool};
			my $volume = $data{devices}{device}{$deviceName}{volume};
			my $message = $data{devices}{device}{$deviceName}{lastmessage};
			
			my $readOnlyStart = ($readonly) ? "<I>" : "";
			my $readOnlyStop = ($readonly) ? "</I>" : "";
			print "<TD BGCOLOR=#C9C9C9>$readOnlyStart<b>$deviceName</b>$readOnlyStop</TD>\n";
			print "<TD BGCOLOR=#C9C9C9>$type</TD>\n";
			print "<TD BGCOLOR=#C9C9C9>$volume</TD>\n";
			print "<TD BGCOLOR=#C9C9C9>($pool)</TD>\n";
			print "</TR>";
			print "<TR>\n";
			print "<TD COLSPAN=4 BGCOLOR=#FAFAFA>$indent $message</TD>";
			print "</TR>\n";
		}
		print "</TABLE>\n";
	}
	print "</DIV>\n";	# Devices
	
	print "<DIV CLASS=messages>\n";
	print "<H2>Messages</H2>\n";
	if (%data && defined($data{messages}) && defined($data{messages}{count}) &&
		$data{messages}{count} > 0) {
		print "<TABLE WIDTH=100% BORDER=0>\n";
		print "<TR>\n";
		print "<TD class=title>Category</td><TD class=title>Priority</TD><TD class=title>Source</TD><TD class=title>Timestamp</TD><TD class=title>Message</TD>\n";
		print "</TR>\n";
		
		my $bgColor = "#C9C9C9";
		
		foreach my $messageID (sort {$b cmp $a} keys $data{messages}{message}) {
			print "<TR>\n";
			my $category = ucfirst($data{messages}{message}{$messageID}{category});
			my $priority = ucfirst($data{messages}{message}{$messageID}{priority});
			my $source = ucfirst($data{messages}{message}{$messageID}{source});
			my $timestamp = $data{messages}{message}{$messageID}{timestamp};
			my $message = ucfirst($data{messages}{message}{$messageID}{message});
			print "<TD BGCOLOR=$bgColor>$category</TD><TD BGCOLOR=$bgColor>$priority</TD><TD BGCOLOR=$bgColor>$source</TD><TD BGCOLOR=$bgColor>$timestamp</TD><TD BGCOLOR=$bgColor>$message</TD>\n";
			print "</TR>\n";
			
			if ($bgColor eq "#C9C9C9") {
				$bgColor = "#FAFAFA";
			} else {
				$bgColor = "#C9C9C9";
			}
		}
		print "</TABLE>\n";
	} else {
		print "<I>No server messages.</I>\n";
	}
	print "</DIV>\n";	# Messages
	
	print "<DIV CLASS=alerts>\n";
	print "<H2>Alerts</H2>\n";
	if (%data && defined($data{alerts}) && defined($data{alerts}{count}) &&
		$data{alerts}{count} > 0) {
		
		print "<TABLE WIDTH=100% BORDER=0>\n";
		print "<TR>\n";
		print "<TD class=title>Level</TD><TD class=title>Timestamp</TD><TD class=title>Message</TD>\n";
		print "</TR>\n";
		my $bgColor = "#C9C9C9";
		
		foreach my $alertID (sort {$a <=> $b} keys $data{alerts}{alert}) {
			print "<TR>\n";
			my $level = ucfirst($data{alerts}{alert}{$alertID}{priority}) . ":" . ucfirst($data{alerts}{alert}{$alertID}{category});
			my $message = $data{alerts}{alert}{$alertID}{message};
			my $timestamp = $data{alerts}{alert}{$alertID}{timestamp};
			print "<TD BGCOLOR=$bgColor><B>$level</B></TD>\n";
			print "<TD BGCOLOR=$bgColor>$timestamp</TD>\n";
			print "<TD BGCOLOR=$bgColor>$message</TD>\n";
						
			if ($bgColor eq "#C9C9C9") {
				$bgColor = "#FAFAFA";
			} else {
				$bgColor = "#C9C9C9";
			}
			print "</TR>\n";
		}
		print "</TABLE>\n";
	} else {
		print "<I>No active alerts.</I>\n";
	}
	print "</DIV>\n";	# Alerts	
	print "</DIV>\n";	# All
}


#########################################################################
# Main
#########################################################################

my $query = new CGI;
print $query->header(-Refresh => "20;url=/cgi-bin/webwatch_2.cgi");
print $query->start_html(
		-title => 'Prototype NSR Webwatch',
		-author => 'preston@nsrd.info',
		-style => { 'src' => '/webwatch_2.css' }
	);
	
print "<DIV CLASS=title>Prototype NSR Webwatch (" . localtime(time) . ")\n</DIV>\n";

get_data();
display_data();

print $query->end_html();
die "\n";
